@@ -1,5 +1,5 @@
 /*
- * Copyright © 2015-2021 Inria.  All rights reserved.
+ * Copyright © 2015-2025 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -49,10 +49,11 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
   unsigned highest_cpuid, highest_ext_cpuid;
   unsigned i;
   int has_intel_x2apic = 0;
+  int has_intel_pconfig = 0;
   int has_intel_sgx = 0;
   int has_amd_topoext = 0;
   FILE *output;
-  int err;
+  int is_amd, err;
 
   err = hwloc_set_cpubind(topo, pu->cpuset, HWLOC_CPUBIND_PROCESS);
   if (err < 0) {
@@ -90,6 +91,15 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
   regs[0] = 0x0;
   dump_one_cpuid(output, regs, 0x1);
 
+  /* AuthenticAMD */
+#define AMD_EBX ('A' | ('u'<<8) | ('t'<<16) | ('h'<<24))
+#define AMD_EDX ('e' | ('n'<<8) | ('t'<<16) | ('i'<<24))
+#define AMD_ECX ('c' | ('A'<<8) | ('M'<<16) | ('D'<<24))
+  if (regs[1] == AMD_EBX && regs[2] == AMD_ECX && regs[3] == AMD_EDX)
+    is_amd = 1;
+  else
+    is_amd = 0;
+
   /* 0x1 = Family, Model, Stepping, Topology, Features */
   if (highest_cpuid >= 0x1) {
     regs[0] = 0x1;
@@ -112,13 +122,15 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
 
   /* 0x4 = Caches on Intel ; Reserved on AMD */
   if (highest_cpuid >= 0x4) {
-    for(i=0; ; i++) {
+    for(i=0; i<256; i++) {
       regs[0] = 0x4; regs[2] = i;
       dump_one_cpuid(output, regs, 0x5);
       if (!(regs[0] & 0x1f))
 	/* invalid, no more caches */
 	break;
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
   }
 
   /* 0x5 = Monitor/mwait */
@@ -138,13 +150,17 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     unsigned max;
     regs[0] = 0x7; regs[2] = 0;
     dump_one_cpuid(output, regs, 0x5);
+    if (regs[3] & (1<<18))
+      has_intel_pconfig = 1;
     if (regs[1] & (1<<2))
       has_intel_sgx = 1;
     max = regs[0];
-    for(i=1; i<=max; i++) {
+    for(i=1; i<=max && i<256; i++) {
       regs[0] = 0x7; regs[2] = i;
       dump_one_cpuid(output, regs, 0x5);
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256 even if max=%u\n", max);
   }
 
   /* 0x9 = DCA on Intel ; Reserved on AMD */
@@ -159,15 +175,23 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     dump_one_cpuid(output, regs, 0x1);
   }
 
-  /* 0xb = Extended topology on Intel ; Reserved on AMD */
+  /* 0xb = Extended Topology Enumeration */
   if (has_intel_x2apic && highest_cpuid >= 0xb) {
-    for(i=0; ; i++) {
+    for(i=0; i<256; i++) {
       regs[0] = 0xb; regs[2] = i;
       dump_one_cpuid(output, regs, 0x5);
-      if (!(regs[2] & 0xff00))
-	/* invalid, no more levels */
-	break;
+      if (is_amd) {
+        /* AMD invalid (no more levels) = logprocatthislevel = ebx[0:15] == 0 */
+        if (!(regs[1] & 0xffff))
+          break;
+      } else {
+        /* Intel invalid (no more levels) = invalid level type = ecx[8:15] == 0 */
+        if (!(regs[2] & 0xff00))
+          break;
+      }
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
   }
 
   /* 0xd = Extended state enumeration */
@@ -212,10 +236,10 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     /* L3 Cache Allocation Technology */
     regs[0] = 0x10; regs[2] = 1;
     dump_one_cpuid(output, regs, 0x5);
-    /* L2 Cache Allocation Technology */
+    /* L2 Cache Allocation Technology on Intel */
     regs[0] = 0x10; regs[2] = 2;
     dump_one_cpuid(output, regs, 0x5);
-    /* Memory Bandwidth Allocation */
+    /* Memory Bandwidth Allocation on Intel */
     regs[0] = 0x10; regs[2] = 3;
     dump_one_cpuid(output, regs, 0x5);
   }
@@ -226,13 +250,15 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     dump_one_cpuid(output, regs, 0x5);
     regs[0] = 0x12; regs[2] = 1;
     dump_one_cpuid(output, regs, 0x5);
-    for(i=2; ; i++) {
+    for(i=2; i<256; i++) {
       regs[0] = 0x12; regs[2] = i;
       dump_one_cpuid(output, regs, 0x5);
       if (!(regs[0] & 0xf))
 	/* invalid, no more subleaves */
 	break;
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
   }
 
   /* 0x14 = Processor trace enumeration on Intel ; Reserved on AMD */
@@ -262,10 +288,12 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     dump_one_cpuid(output, regs, 0x5);
     maxsocid = regs[0];
     if (maxsocid >= 3) {
-      for(i=1; i<=maxsocid; i++) {
+      for(i=1; i<=maxsocid && i<256; i++) {
 	regs[0] = 0x17; regs[2] = i;
 	dump_one_cpuid(output, regs, 0x5);
       }
+      if (i == 256)
+        fprintf(output, "# stopped at ecx=256 even if maxsocid=%u\n", maxsocid);
     }
   }
 
@@ -275,7 +303,7 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     regs[0] = 0x18; regs[2] = 0;
     dump_one_cpuid(output, regs, 0x5);
     max = regs[0];
-    for(i=1; i<=max; i++) {
+    for(i=1; i<=max && i<256; i++) {
       regs[0] = 0x18; regs[2] = i;
       regs[3] = 0; /* mark as invalid in case the cpuid call doesn't do anything */
       dump_one_cpuid(output, regs, 0x5);
@@ -283,6 +311,8 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
 	/* invalid, but it doesn't mean the next subleaf will be invalid */
         continue;
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256 even if max=%u\n", max);
   }
 
   /* 0x19 = Key Locker Leaf on Intel ; Reserved on AMD */
@@ -297,20 +327,90 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     dump_one_cpuid(output, regs, 0x5);
   }
 
-  /* 0x1b = (Removed) PCONFIG Information on Intel ; Reserved on AMD */
+  /* 0x1b = PCONFIG Information on Intel ; Reserved on AMD */
+  if (has_intel_pconfig && highest_cpuid >= 0x1b) {
+    for(i=0; i<256; i++) {
+      regs[0] = 0x1b; regs[2] = i;
+      dump_one_cpuid(output, regs, 0x5);
+      if (!(regs[0] & 0xfff))
+	break;
+    }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
+  }
+
+  /* 0x1c = Last Branch Records Information on Intel ; Reserved on AMD */
+  if (highest_cpuid >= 0x1c) {
+    regs[0] = 0x1c; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+  }
+
+  /* 0x1d = Tile Information ; Reserved on AMD */
+  if (highest_cpuid >= 0x1d) {
+    regs[0] = 0x1d; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+    regs[0] = 0x1d; regs[2] = 1;
+    dump_one_cpuid(output, regs, 0x5);
+  }
+
+  /* 0x1e = TMUL Information ; Reserved on AMD */
+  if (highest_cpuid >= 0x1e) {
+    regs[0] = 0x1e; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+  }
 
   /* 0x1f = V2 Extended Topology Enumeration on Intel ; Reserved on AMD */
   if (highest_cpuid >= 0x1f) {
-    for(i=0; ; i++) {
+    for(i=0; i<256; i++) {
       regs[0] = 0x1f; regs[2] = i;
       dump_one_cpuid(output, regs, 0x5);
       if (!(regs[2] & 0xff00))
 	/* invalid, no more levels */
 	break;
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
   }
 
-  if (highest_cpuid > 0x1f) {
+  /* 0x20 = Processor History Reset on Intel ; Reserved on AMD */
+  if (highest_cpuid >= 0x20) {
+    regs[0] = 0x20; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+    /* eax is number of subleaves but subleaves aren't documented?! */
+  }
+
+  /* 0x21 is Unimplemented on Intel (always returns 0) ; Reserved on AMD */
+  if (highest_cpuid >= 0x21) {
+    regs[0] = 0x21; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+  }
+
+  /* TODO 0x22 is reserved on Intel and AMD? */
+  if (highest_cpuid >= 0x22) {
+    regs[0] = 0x22; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+  }
+
+  /* 0x23 is Architectural Performance Monitoring Extended Leaf on Intel ; Reserved on AMD */
+  if (highest_cpuid >= 0x23) {
+    unsigned subleafmask;
+    regs[0] = 0x23; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+    subleafmask = regs[0];
+    for(i=1; i<32; i++)
+      if ((1U<<i) & subleafmask) {
+        regs[0] = 0x23; regs[2] = i;
+        dump_one_cpuid(output, regs, 0x5);
+      }
+  }
+
+  /* 0x24 is Converged Vector ISA Main Leaf on Intel ; Reserved on AMD */
+  if (highest_cpuid >= 0x24) {
+    regs[0] = 0x24; regs[2] = 0;
+    dump_one_cpuid(output, regs, 0x5);
+  }
+
+  if (highest_cpuid > 0x25) {
     static int reported = 0;
     if (!reported)
       fprintf(stderr, "WARNING: Processor supports new CPUID leaves upto 0x%x\n", highest_cpuid);
@@ -399,13 +499,15 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
 
   /* 0x8000001d = Cache properties on AMD ; Reserved on Intel */
   if (highest_ext_cpuid >= 0x8000001d) {
-    for(i=0; ; i++) {
+    for(i=0; i<256; i++) {
       regs[0] = 0x8000001d; regs[2] = i;
       dump_one_cpuid(output, regs, 0x5);
       if (!(regs[0] & 0x1f))
 	/* no such cache, no more cache */
 	break;
     }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
   }
 
   /* 0x8000001e = Topoext on AMD ; Reserved on Intel */
@@ -414,13 +516,13 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     dump_one_cpuid(output, regs, 0x1);
   }
 
-  /* 0x8000001f = Encrypted Memory Capabilities ; Reserved on Intel */
+  /* 0x8000001f = Encrypted Memory Capabilities on AMD ; Reserved on Intel */
   if (highest_ext_cpuid >= 0x8000001f) {
     regs[0] = 0x8000001f;
     dump_one_cpuid(output, regs, 0x1);
   }
 
-  /* 0x80000020 = Platform QoS Enforcement for Memory Bandwidth */
+  /* 0x80000020 = Platform QoS Enforcement for Memory Bandwidth on AMD ; Reserved on Intel */
   if (highest_ext_cpuid >= 0x80000020) {
     regs[0] = 0x80000020; regs[2] = 0;
     dump_one_cpuid(output, regs, 0x5);
@@ -428,7 +530,40 @@ static int dump_one_proc(hwloc_topology_t topo, hwloc_obj_t pu, const char *path
     dump_one_cpuid(output, regs, 0x5);
   }
 
-  if (highest_ext_cpuid > 0x8000001f) {
+  /* 0x80000021 = Extended Feature Identification 2 on AMD ; Reserved on Intel */
+  if (highest_ext_cpuid >= 0x80000021) {
+    regs[0] = 0x80000021;
+    dump_one_cpuid(output, regs, 0x1);
+  }
+
+  /* 0x80000022 = Extended Performance Monitoring and Debug on AMD ; Reserved on Intel */
+  if (highest_ext_cpuid >= 0x80000022) {
+    regs[0] = 0x80000022;
+    dump_one_cpuid(output, regs, 0x1);
+  }
+
+  /* 0x80000023 = Secure Multi-Key Encryption on AMD ; Reserved on Intel */
+  if (highest_ext_cpuid >= 0x80000023) {
+    regs[0] = 0x80000023;
+    dump_one_cpuid(output, regs, 0x1);
+  }
+
+  /* 0x80000026 = Extended CPU Topology on AMD ; Reserved on Intel */
+  if (highest_ext_cpuid >= 0x80000026) {
+    for(i=0; i<256; i++) {
+      regs[0] = 0x80000026; regs[2] = i;
+      dump_one_cpuid(output, regs, 0x5);
+      if (!(regs[1] & 0xffff))
+	/* nothing here, no more levels */
+	break;
+    }
+    if (i == 256)
+      fprintf(output, "# stopped at ecx=256\n");
+  }
+
+  /* AMD Genoa reports highest_ext_cpuid == 0x80000028 but these last 2 leaves aren't documented yet */
+
+  if (highest_ext_cpuid > 0x80000026) {
     static int reported = 0;
     if (!reported)
       fprintf(stderr, "WARNING: Processor supports new extended CPUID leaves upto 0x%x\n", highest_ext_cpuid);
@@ -445,9 +580,10 @@ void usage(const char *callname, FILE *where)
   fprintf(where, "Usage : %s [ options ] ... [ outdir ]\n", callname);
   fprintf(where, "  outdir is an optional output directory instead of cpuid/\n");
   fprintf(where, "Options:\n");
-  fprintf(where, "  -c <n>       Only gather for logical processor with logical index <n>\n");
-  fprintf(where, "  -s --silent  Do not show verbose messages\n");
-  fprintf(where, "  -h --help    Show this usage\n");
+  fprintf(where, "  -c <n>          Only gather for logical processor with logical index <n>\n");
+  fprintf(where, "  -q --quiet -s   Do not show verbose messages\n");
+  fprintf(where, "  --version       Report version and exit\n");
+  fprintf(where, "  -h --help       Show this usage\n");
 }
 
 int main(int argc, const char * const argv[])
@@ -462,7 +598,12 @@ int main(int argc, const char * const argv[])
   int err;
   int ret = EXIT_SUCCESS;
 
-  callname = argv[0];
+  callname = strrchr(argv[0], '/');
+  if (!callname)
+    callname = argv[0];
+  else
+    callname++;
+
   argc--; argv++;
 
   hwloc_utils_check_api_version(callname);
@@ -478,10 +619,14 @@ int main(int argc, const char * const argv[])
       idx = atoi(argv[1]);
       argc -= 2;
       argv += 2;
-    } else if (argc >= 1 && (!strcmp(argv[0], "-s") || !strcmp(argv[0], "--silent"))) {
+    } else if (argc >= 1 && (!strcmp(argv[0], "-q") || !strcmp(argv[0], "--quiet")
+                             || !strcmp(argv[0], "-s") || !strcmp(argv[0], "--silent"))) {
       verbose--;
       argc--;
       argv++;
+    } else if (!strcmp (argv[0], "--version")) {
+      printf("%s %s\n", callname, HWLOC_VERSION);
+      exit(EXIT_SUCCESS);
     } else if (!strcmp(argv[0], "-h") || !strcmp(argv[0], "--help")) {
       usage(callname, stdout);
       goto out;
@@ -501,6 +646,9 @@ int main(int argc, const char * const argv[])
 
   hwloc_topology_init(&topo);
   hwloc_topology_set_all_types_filter(topo, HWLOC_TYPE_FILTER_KEEP_NONE);
+  hwloc_topology_set_flags(topo, HWLOC_TOPOLOGY_FLAG_NO_DISTANCES
+                                 |HWLOC_TOPOLOGY_FLAG_NO_MEMATTRS
+                                 |HWLOC_TOPOLOGY_FLAG_NO_CPUKINDS);
   err = hwloc_topology_load(topo);
   if (err < 0) {
     fprintf(stderr, "Failed to load topology\n");
